import signal
import time
import unittest

from plop import platform

class WorkaroundTests(unittest.TestCase):
    def test_seconds_to_timeval(self):
        def tup(tv): return (tv.tv_sec, tv.tv_usec)
        self.assertEqual(tup(platform.seconds_to_timeval(1)), (1, 0))
        self.assertEqual(tup(platform.seconds_to_timeval(2.5)), (2, 500000))
        # We lose a little precision at the low end.  Oh well, we're probably
        # not getting microsecond accuracy in python anyway.
        self.assertEqual(tup(platform.seconds_to_timeval(1.000010)), (1, 10))

if hasattr(signal, 'setitimer'):
    del WorkaroundTests

class ItimerTests(unittest.TestCase):
    def test_timer(self):
        result = []
        def handler(sig, frame):
            result.append(None)  # peano arithmetic ftw
        signal.signal(signal.SIGPROF, handler)
        platform.setitimer(platform.ITIMER_PROF, 0.01, 0.01)
        start = time.time()
        # must busy-wait because if we're sleeping python signal handlers
        # don't run.
        while time.time() < start + 0.5: pass
        platform.setitimer(platform.ITIMER_PROF, 0, 0)
        end = time.time()
        elapsed = end - start
        self.assertTrue(0.500 < elapsed < 0.501, elapsed)
        count = len(result)
        # We should have gotten ~50 timer ticks while we waited.
        # in practice I see ~44 on my mac, and ~48 on a linux
        # vm running on that same mac (?!)
        self.assertTrue(37 < count < 55, count)



import ast
import logging
import threading
import time
import unittest
import six

from plop.collector import Collector, PlopFormatter

class CollectorTest(unittest.TestCase):
    def filter_stacks(self, collector):
        # Kind of hacky, but this is the simplest way to keep the tests
        # working after the internals of the collector changed to support
        # multiple formatters.
        stack_counts = ast.literal_eval(PlopFormatter().format(collector))
        counts = {}
        for stack, count in six.iteritems(stack_counts):
            filtered_stack = [frame[2] for frame in stack
                              if frame[0].endswith('collector_test.py')]
            if filtered_stack:
                counts[tuple(filtered_stack)] = count
        return counts
        
    def check_counts(self, counts, expected):
        failed = False
        output = []
        for stack, count in six.iteritems(expected):
            # every expected frame should appear in the data, but
            # the inverse is not true if the signal catches us between
            # calls.
            self.assertTrue(stack in counts)
            ratio = float(counts[stack])/float(count)
            output.append('%s: expected %s, got %s (%s)' % 
                          (stack, count, counts[stack], ratio))
            if not (0.70 <= ratio <= 1.25):
                failed = True
        if failed:
            for line in output:
                logging.warning(line)
            for key in set(counts.keys()) - set(expected.keys()):
                logging.warning('unexpected key: %s: got %s' % (key, counts[key]))
            self.fail("collected data did not meet expectations")

    def test_collector(self):
        start = time.time()
        def a(end):
            while time.time() < end: pass
            c(time.time() + 0.1)
        def b(end):
            while time.time() < end: pass
            c(time.time() + 0.1)
        def c(end):
            while time.time() < end: pass
        collector = Collector(interval=0.01, mode='prof')
        collector.start()
        a(time.time() + 0.1)
        b(time.time() + 0.2)
        c(time.time() + 0.3)
        end = time.time()
        collector.stop()
        elapsed = end - start
        self.assertTrue(0.8 < elapsed < 0.9, elapsed)

        counts = self.filter_stacks(collector)
        
        expected = {
            ('a', 'test_collector'): 10,
            ('c', 'a', 'test_collector'): 10,
            ('b', 'test_collector'): 20,
            ('c', 'b', 'test_collector'): 10,
            ('c', 'test_collector'): 30,
            }
        self.check_counts(counts, expected)

        # cost depends on stack depth; for this tiny test I see 40-80usec
        time_per_sample = float(collector.sample_time) / collector.samples_taken
        self.assertTrue(time_per_sample < 0.000100, time_per_sample)

    # TODO: any way to make this test not flaky?
    def disabled_test_collect_threads(self):
        start = time.time()
        def a(end):
            while time.time() < end: pass
        def thread1_func():
            a(time.time() + 0.2)
        def thread2_func():
            a(time.time() + 0.3)
        collector = Collector(interval=0.01, mode='prof')
        collector.start()
        thread1 = threading.Thread(target=thread1_func)
        thread2 = threading.Thread(target=thread2_func)
        thread1.start()
        thread2.start()
        a(time.time() + 0.1)
        while thread1.isAlive(): pass
        while thread2.isAlive(): pass
        thread1.join()
        thread2.join()
        end = time.time()
        collector.stop()
        elapsed = end - start
        self.assertTrue(0.3 < elapsed < 0.4, elapsed)

        counts = self.filter_stacks(collector)

        expected = {
            ('a', 'test_collect_threads'): 10,
            ('a', 'thread1_func'): 20,
            ('a', 'thread2_func'): 30,
            }
        self.check_counts(counts, expected)



#!/usr/bin/env python

import sys
import unittest

COMMON_TESTS = [
    'plop.test.collector_test',
    'plop.test.platform_test',
    ]

VIEWER_TESTS = [
    'plop.test.callgraph_test',
    ]

# Viewer currently requires python 2.7 (for collections.Counter, could
# probably be fixed); collector supports >= 2.5.
if sys.version_info[:2] >= (2,7):
    TEST_MODULES = COMMON_TESTS + VIEWER_TESTS
else:
    TEST_MODULES = COMMON_TESTS

def all():
    return unittest.defaultTestLoader.loadTestsFromNames(TEST_MODULES)

if __name__ == '__main__':
    import tornado.testing
    tornado.testing.main()



import logging
import unittest

from plop.callgraph import CallGraph, Node

class SimpleCallgraphTest(unittest.TestCase):
    def setUp(self):
        graph = CallGraph()
        graph.add_stack([Node(1), Node(2)], dict(time=1))
        graph.add_stack([Node(1), Node(3)], dict(time=3))
        graph.add_stack([Node(1), Node(2), Node(3)], dict(time=7))
        graph.add_stack([Node(1), Node(4), Node(2), Node(3)], dict(time=2))
        self.graph = graph

    def test_basic_attrs(self):
        logging.debug(self.graph.nodes)
        logging.debug(self.graph.edges)
        self.assertEqual(len(self.graph.nodes), 4)
        self.assertEqual(len(self.graph.edges), 5)

    def test_top_edges(self):
        top_edges = self.graph.get_top_edges('time', 3)
        logging.debug(top_edges)
        summary = [(e.parent.id, e.child.id, e.weights['time']) for e in top_edges]
        self.assertEqual(summary, [
                (2, 3, 9),
                (1, 2, 8),
                (1, 3, 3),
                ])

    def test_top_nodes(self):
        top_nodes = self.graph.get_top_nodes('time', 2)
        logging.debug(top_nodes)
        summary = [(n.id, n.weights['time']) for n in top_nodes]
        self.assertEqual(summary, [
                (3, 12),
                (2, 1),
                ])



from collections import Counter
import six


class Node(object):
    def __init__(self, id, attrs=None):
        self.id = id
        self.attrs = attrs or {}
        self.weights = Counter()

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return 'Node(%s)' % (self.id,)

class Edge(object):
    def __init__(self, parent, child, weights):
        self.parent = parent
        self.child = child
        self.weights = Counter(weights)

    def key(self):
        return (self.parent, self.child)

    def __hash__(self):
        return hash(self.key())

    def __eq__(self, other):
        return self.key() == other.key()

    def __repr__(self):
        return 'Edge(%s, %s, %r)' % (self.parent.id, self.child.id, self.weights)

    def __iadd__(self, other):
        self.weights += other.weights
        return self

class Stack(object):
    def __init__(self, nodes, weights):
        self.nodes = nodes
        self.weights = Counter(weights)


class CallGraph(object):
    def __init__(self):
        # map Node.id: Node
        self.nodes = {}
        self.edges = {}
        self.stacks = []

    def add_stack(self, nodes, weights):
        nodes = [self.nodes.setdefault(n.id, n) for n in nodes]
        weights = Counter(weights)
        self.stacks.append(Stack(nodes, weights))
        
        for i in range(len(nodes) - 1):
            parent = nodes[i]
            child = nodes[i + 1]
            edge = Edge(parent, child, weights)
            if edge.key() in self.edges:
                self.edges[edge.key()] += edge
            else:
                self.edges[edge.key()] = edge
        nodes[-1].weights += weights

    def get_top_edges(self, weight, num=10):
        # TODO: is there a partial sort in python?
        sorted_edges = sorted(six.itervalues(self.edges),
                              key=lambda e: e.weights.get(weight, 0),
                              reverse=True)
        return sorted_edges[:num]

    def get_top_nodes(self, weight, num=10):
        sorted_nodes = sorted(six.itervalues(self.nodes),
                              key=lambda n: n.weights.get(weight, 0),
                              reverse=True)
        return sorted_nodes[:num]

    @staticmethod
    def load(filename):
        import ast
        with open(filename) as f:
            data = ast.literal_eval(f.read())
        graph = CallGraph()
        for stack, count in six.iteritems(data):
            stack_nodes = [Node(id=frame, attrs=dict(fullpath=frame[0],
                                                     filename=frame[0].rpartition('/')[-1],
                                                     lineno=frame[1],
                                                     funcname=frame[2]))
                           for frame in reversed(stack)]
            # TODO: this shouldn't be recorded as "calls"
            graph.add_stack(stack_nodes, weights=dict(calls=count))
        return graph

if __name__ == '__main__':
    import sys
    graph = CallGraph.load(sys.argv[1])



from __future__ import with_statement
import collections
import errno
import os
import signal
import sys
from six.moves import _thread
import time
import argparse
import six
import plop.platform


class Collector(object):
    MODES = {
        'prof': (plop.platform.ITIMER_PROF, signal.SIGPROF),
        'virtual': (plop.platform.ITIMER_VIRTUAL, signal.SIGVTALRM),
        'real': (plop.platform.ITIMER_REAL, signal.SIGALRM),
    }

    def __init__(self, interval=0.01, mode='virtual'):
        self.interval = interval
        self.mode = mode
        assert mode in Collector.MODES
        timer, sig = Collector.MODES[self.mode]
        signal.signal(sig, self.handler)
        signal.siginterrupt(sig, False)
        self.reset()

    def reset(self):
        self.stacks = list()
        self.samples_remaining = 0
        self.stopping = False
        self.stopped = False

        self.samples_taken = 0
        self.sample_time = 0

    def start(self, duration=30.0):
        self.stopping = False
        self.stopped = False
        self.samples_remaining = int(duration / self.interval)
        timer, sig = Collector.MODES[self.mode]
        plop.platform.setitimer(timer, self.interval, self.interval)

    def stop(self):
        self.stopping = True
        self.wait()

    def wait(self):
        while not self.stopped:
            pass  # need busy wait; ITIMER_PROF doesn't proceed while sleeping

    def handler(self, sig, current_frame):
        start = time.time()
        self.samples_remaining -= 1
        if self.samples_remaining <= 0 or self.stopping:
            plop.platform.setitimer(Collector.MODES[self.mode][0], 0, 0)
            self.stopped = True
            return
        current_tid = _thread.get_ident()
        for tid, frame in six.iteritems(sys._current_frames()):
            if tid == current_tid:
                frame = current_frame
            frames = []
            while frame is not None:
                code = frame.f_code
                frames.append((code.co_filename, frame.f_lineno, code.co_name))
                frame = frame.f_back
            self.stacks.append(frames)
        end = time.time()
        self.samples_taken += 1
        self.sample_time += (end - start)


class CollectorFormatter(object):
    """
    Abstract class for output formats
    """
    def format(self, collector):
        raise Exception("not implemented")

    def store(self, collector, filename):
        with open(filename, "wb") as f:
            f.write(self.format(collector).encode('utf-8'))


class PlopFormatter(CollectorFormatter):
    """
    Formats stack frames for plop.viewer
    """
    def __init__(self, max_stacks=50):
        self.max_stacks = 50

    def format(self, collector):
        # defaultdict instead of counter for pre-2.7 compatibility
        stack_counts = collections.defaultdict(int)
        for frames in collector.stacks:
            stack_counts[tuple(frames)] += 1
        stack_counts = dict(sorted(six.iteritems(stack_counts),
                                   key=lambda kv: -kv[1])[:self.max_stacks])
        return repr(stack_counts)


class FlamegraphFormatter(CollectorFormatter):
    """
    Creates Flamegraph files
    """
    def format(self, collector):
        output = ""
        previous = None
        previous_count = 1
        for stack in collector.stacks:
            current = self.format_flame(stack)
            if current == previous:
                previous_count += 1
            else:
                output += "%s %d\n" % (previous, previous_count)
                previous_count = 1
                previous = current
        output += "%s %d\n" % (previous, previous_count)
        return output

    def format_flame(self, stack):
        funcs = map("{0[2]} ({0[0]}:{0[1]})".format, reversed(stack))
        return ";".join(funcs)


def main():
    # TODO: more options, refactor this into somewhere shared
    # between tornado.autoreload and auto2to3
    parser = argparse.ArgumentParser(description="Plop: Python Low-Overhead Profiler",
                                     prog="python -m plop.collector",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--format", "-f", help="Output format",
                        choices=["plop", "flamegraph"], default="plop")
    parser.add_argument("--module", "-m", help="Execute target as a module",
                        action="store_const", const=True, default=False)
    parser.add_argument("--mode", help="Interval timer mode to use, see `man 2 setitimer`",
                        choices=["prof", "real", "virtual"], default="prof")
    parser.add_argument("--interval", help="Timer interval in seconds", default=0.01, type=float)
    parser.add_argument("--duration", help="Profiling duration in seconds", default=3600,
                        type=int)
    parser.add_argument("--max-stacks", help=("Number of most frequent stacks to store."
                        " Ignored for Flamegraph output."), type=int, default=50)
    parser.add_argument("target", help="Module or script to run")
    parser.add_argument("arguments", nargs=argparse.REMAINDER,
                        help="Pass-through arguments for the profiled application")
    args = parser.parse_args()
    sys.argv = [args.target] + args.arguments

    if args.format == "flamegraph":
        extension = "flame"
        formatter = FlamegraphFormatter()
    elif args.format == "plop":
        extension = "plop"
        formatter = PlopFormatter(max_stacks=args.max_stacks)
    else:
        sys.stderr.write("Unhandled output format: %s" % args.format)
        sys.stderr.flush()
        sys.exit(1)

    try:  # This is equivalent to os.makedirs('profiles', exist_ok=True)
        os.mkdir('profiles')
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
    filename = os.path.join(
        'profiles',
        '%s-%s.%s' % (os.path.basename(args.target),
                      time.strftime('%Y%m%d-%H%M-%S'),
                      extension))

    collector = Collector(mode=args.mode, interval=args.interval)
    collector.start(duration=args.duration)
    exit_code = 0
    try:
        if args.module:
            import runpy
            runpy.run_module(args.target, run_name="__main__", alter_sys=True)
        else:
            with open(args.target) as f:
                # Execute the script in our namespace instead of creating
                # a new one so that something that tries to import __main__
                # (e.g. the unittest module) will see names defined in the
                # script instead of just those defined in this module.
                global __file__
                __file__ = args.target
                # If __package__ is defined, imports may be incorrectly
                # interpreted as relative to this module.
                global __package__
                del __package__
                six.exec_(f.read(), globals(), globals())
    except SystemExit as e:
        exit_code = e.code
    collector.stop()
    if collector.samples_taken:
        formatter.store(collector, filename)
        print("profile output saved to %s" % filename)
        overhead = float(collector.sample_time) / collector.samples_taken
        print("overhead was %s per sample (%s%%)" % (
            overhead, overhead / collector.interval))
    else:
        print("no samples collected; program was too fast")
    sys.exit(exit_code)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from collections import Counter
import logging
import os

from tornado.ioloop import IOLoop
from tornado.options import define, options, parse_command_line
from tornado.web import RequestHandler, Application
import six

from plop.callgraph import CallGraph

define('port', default=8888)
define('debug', default=False)
define('address', default='')
define('datadir', default='profiles')

class IndexHandler(RequestHandler):
    def get(self):
        files = []
        for filename in os.listdir(options.datadir):
            mtime = os.stat(os.path.join(options.datadir, filename)).st_mtime
            files.append((mtime, filename))
        # sort by descending mtime then ascending filename
        files.sort(key=lambda x: (-x[0], x[1]))
        self.render('index.html', files=[f[1] for f in files])

class ViewHandler(RequestHandler):
    def get(self):
        self.render('force.html', filename=self.get_argument("filename"))

class ViewFlatHandler(RequestHandler):
    def get(self):
        self.render('force-flat.html',
                    data=profile_to_json(self.get_argument('filename')))

    def embed_file(self, filename):
        with open(os.path.join(self.settings['static_path'], filename)) as f:
            return f.read()

class DataHandler(RequestHandler):
    def get(self):
        self.write(profile_to_json(self.get_argument('filename')))

def profile_to_json(filename):
    root = os.path.abspath(options.datadir) + os.path.sep
    abspath = os.path.abspath(os.path.join(root, filename))
    assert (abspath + os.path.sep).startswith(root)
    graph = CallGraph.load(abspath)

    total = sum(stack.weights['calls'] for stack in graph.stacks)
    top_stacks = graph.stacks
    #top_stacks = [stack for stack in graph.stacks if stack.weights['calls'] > total*.005]
    filtered_nodes = set()
    for stack in top_stacks:
        filtered_nodes.update(stack.nodes)
    nodes=[dict(attrs=node.attrs, weights=node.weights, id=node.id)
           for node in filtered_nodes]
    nodes = sorted(nodes, key=lambda n: -n['weights']['calls'])
    #index = {node['id']: i for i, node in enumerate(nodes)}
    index = dict([(node['id'], i) for i, node in enumerate(nodes)])


    # High-degree nodes are generally common utility functions, and
    # creating edges from all over the graph tends to obscure more than
    # it helps.
    degrees = Counter()
    dropped = set()
    for edge in six.itervalues(graph.edges):
        degrees[edge.child.id] += 1
        degrees[edge.parent.id] += 1
    for node, degree in six.iteritems(degrees):
        if degree > 6:
            dropped.add(node)

    edges = [dict(source=index[edge.parent.id],
                  target=index[edge.child.id],
                  weights=edge.weights)
             for edge in six.itervalues(graph.edges)
             if (edge.parent.id in index and
                 edge.child.id in index and
                 edge.parent.id not in dropped and
                 edge.child.id not in dropped)]
    stacks = [dict(nodes=[index[n.id] for n in stack.nodes],
                   weights=stack.weights)
              for stack in top_stacks]
    return dict(nodes=nodes, edges=edges, stacks=stacks)

def main():
    parse_command_line()

    handlers = [
        ('/', IndexHandler),
        ('/view', ViewHandler),
        ('/view-flat', ViewFlatHandler),
        ('/data', DataHandler),
        ]

    settings=dict(
        debug=options.debug,
        static_path=os.path.join(os.path.dirname(__file__), 'static'),
        template_path=os.path.join(os.path.dirname(__file__), 'templates'),
        )

    app = Application(handlers, **settings)
    app.listen(options.port, address=options.address)
    print("server starting at http://%s:%s" % (options.address or 'localhost',
                                               options.port))
    IOLoop.instance().start()

if __name__ == '__main__':
    main()



import signal

if hasattr(signal, 'setitimer'):
    from signal import setitimer, ITIMER_REAL, ITIMER_VIRTUAL, ITIMER_PROF
else:
    # python2.5 doesn't have setitimer
    import ctypes.util
    libc = ctypes.CDLL(ctypes.util.find_library("c"))
    class Timeval(ctypes.Structure):
        _fields_ = [('tv_sec', ctypes.c_long), ('tv_usec', ctypes.c_long)]
    class Itimerval(ctypes.Structure):
        _fields_ = [('it_interval', Timeval), ('it_value', Timeval)]
    libc.setitimer.argtypes = [ctypes.c_int, ctypes.POINTER(Itimerval),
                               ctypes.POINTER(Itimerval)]
    def seconds_to_timeval(seconds):
        return Timeval(int(seconds), int((seconds % 1) * 1000000))
    def setitimer(which, seconds, interval):
        libc.setitimer(which, 
                       Itimerval(seconds_to_timeval(interval),
                                 seconds_to_timeval(seconds)),
                       # We should get the output value here to return,
                       # but we don't need it right now.
                       None)
    ITIMER_REAL = 0
    ITIMER_VIRTUAL = 1
    ITIMER_PROF = 2
